/*
  the actual renderer
*/

#ifndef __GEZQUADTREERENDERER_H__
#define __GEZQUADTREERENDERER_H__

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "Error.h"
#include "Renderer.h"
#include "Console.h"
#include "Map.h"
#include "Vector.h"
#include "culling.h"
#include <GL\gl.h>
#include <GL\glext.h>

//#define NORMALIZE_PLANES
//#define DRAW_CULLED_NODES /* draw culled nodes in red color for testing... */
#define NODE_UNDEFINED       0x00 // node not visited
#define NODE_RENDER          0x01 // node to render 
#define NODE_REFINE          0x02
#define NODE_CULLED          0x04

#define GL_QUERY_PROC(name, type)           \
	name = (type) wglGetProcAddress(#name); \
	console->Output("querying " #name);     \
	if (name == NULL)                       \
	{                                       \
		error++;                            \
		console->Output(" failed!\n");      \
	}										\
	else									\
	{										\
		console->Output(" ok\n");			\
	}										\
	ASSERT(name != NULL, #name " is NULL");

typedef GLfloat vertex[3];
typedef GLfloat text_coord[2];

class QuadTreeRenderer : public Renderer
{
  private:

	int min_detail, max_detail;

	int MAP_WIDTH, MAP_HEIGHT;

	Plane cull_planes[6];

	vertex *color_data;
	text_coord *texture_coords;

	GLfloat *d_array;

    GLuint index_count,
	       **multi_indexes,
		   *multi_indexes_ARB;
	GLsizei *multi_lengths;

	GLbyte *quadtree_data;

	GLuint texName;

	// EXT multi_draw_array
	bool multi_draw;
	PFNGLMULTIDRAWARRAYSEXTPROC   glMultiDrawArraysEXT;
	PFNGLMULTIDRAWELEMENTSEXTPROC glMultiDrawElementsEXT;


    // ARB extension function pointers...
	PFNGLBINDBUFFERARBPROC           glBindBufferARB;
    PFNGLDELETEBUFFERSARBPROC        glDeleteBuffersARB;
	PFNGLGENBUFFERSARBPROC           glGenBuffersARB;
	PFNGLISBUFFERARBPROC             glIsBufferARB;
	PFNGLBUFFERDATAARBPROC           glBufferDataARB;
	PFNGLBUFFERSUBDATAARBPROC        glBufferSubDataARB;
    PFNGLGETBUFFERSUBDATAARBPROC     glGetBufferSubDataARB;
	PFNGLMAPBUFFERARBPROC            glMapBufferARB;
	PFNGLUNMAPBUFFERARBPROC          glUnmapBufferARB;
	PFNGLGETBUFFERPARAMETERIVARBPROC glGetBufferParameterivARB;
    PFNGLGETBUFFERPOINTERVARBPROC    glGetBufferPointervARB;

	GLuint *index_data,
		   *index_pointer;

    // function prototypes
	void CalculateNormalized_d2(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);
	GLfloat Calculate_d2_Recursion(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);
	void PreventCracks_d2(unsigned int edge_len);

	void QuadTreeTesselate(const unsigned int x1, const unsigned int y1, const unsigned int x2, const unsigned int y2, const Vector3f &vPos, bool cull);


    void QuadTreeRender_TriangleFan_Recursion(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);

	GEZERROR InitImmediateMode(Preferences &prefs);
	void QuadTreeRender_Immediate(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);
	void QuadTreeRender_Immediate_VertexArray(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);

	GEZERROR InitARBExtension(Preferences &prefs);
    void QuadTreeRender_ARB(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);
    void QuadTreeRender_TriangleFan_Recursion_ARB(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);

	typedef void (QuadTreeRenderer::*QUADTREERENDER)(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2);
	QUADTREERENDER QuadTreeRender;

	// texture mapping
	GLuint text_width, text_height;
	GLvoid *text_data;
	bool texture_mapping;

	bool load_texture(char *name);

  protected:
  public:
	/* culling planes */

	float C; // global minimum detail constant
	float c; // local desired detail

	bool frustum_culling,
		 show_cull;

	vertex *vertex_data;

	GLuint fan_current;
	
	// EXT compiled_vertex_array
	bool compiled_array;
	PFNGLLOCKARRAYSEXTPROC glLockArraysEXT;
	PFNGLUNLOCKARRAYSEXTPROC glUnlockArraysEXT;


	QuadTreeRenderer();
    ~QuadTreeRenderer();
    
    GEZERROR Initialize(Console &console, Preferences &params);
    GEZERROR Render(const Vector3f &pos, const Vector3f &rot);
    void CleanUp();

	void IncreaseDetailLevel();
	void DecreaseDetailLevel();
	void IncreaseLocalDetailLevel();
	void DecreaseLocalDetailLevel();
	float GetDetailLevel() const;
	float GetLocalDetailLevel() const;

};

#endif /* ifndef __GEZQUADTREERENDERER_H__ */

