#include "QuadTreeRenderer.h"
#include "glFuncs.h"
#include "GL\glext.h"
#include "error.h"

#define QUAD_VERTEX_BUFFER 1
#define QUAD_INDEX_BUFFER 2

GEZERROR QuadTreeRenderer::InitARBExtension(Preferences &prefs)
{
	int error;

	error = 0;
	console->Output("getting GL_ARB_vertex_array_object ");
	if (QueryExtension("GL_ARB_vertex_buffer_object"))
	{
		console->Output("\n");

		GL_QUERY_PROC(glBindBufferARB,           PFNGLBINDBUFFERARBPROC)
		GL_QUERY_PROC(glDeleteBuffersARB,        PFNGLDELETEBUFFERSARBPROC)
		GL_QUERY_PROC(glGenBuffersARB,           PFNGLGENBUFFERSARBPROC)
		GL_QUERY_PROC(glIsBufferARB,             PFNGLISBUFFERARBPROC)
		GL_QUERY_PROC(glBufferDataARB,           PFNGLBUFFERDATAARBPROC)
		GL_QUERY_PROC(glBufferSubDataARB,        PFNGLBUFFERSUBDATAARBPROC)
		GL_QUERY_PROC(glGetBufferSubDataARB,     PFNGLGETBUFFERSUBDATAARBPROC)
		GL_QUERY_PROC(glMapBufferARB,            PFNGLMAPBUFFERARBPROC)
		GL_QUERY_PROC(glUnmapBufferARB,          PFNGLUNMAPBUFFERARBPROC)
		GL_QUERY_PROC(glGetBufferParameterivARB, PFNGLGETBUFFERPARAMETERIVARBPROC)
		GL_QUERY_PROC(glGetBufferPointervARB,    PFNGLGETBUFFERPOINTERVARBPROC)
	}
	else
	{
		console->Output("not available\n");
		error = 1;
	}

	if (error) return GEZ_UNKNOWN;

	if (multi_indexes_ARB != NULL) delete[] multi_indexes_ARB;
	multi_indexes_ARB = new GLuint[MAP_WIDTH * MAP_HEIGHT];
	ASSERT(multi_indexes_ARB != NULL, "out of memory");
	if (multi_indexes_ARB == NULL)
	{
		return GEZ_UNKNOWN; // TODO: don't fail, we could fall back...
	} // TODO: don't do that if not needed

	if (multi_lengths != NULL) delete[] multi_lengths;
	multi_lengths = new GLsizei[MAP_WIDTH * MAP_HEIGHT];
	ASSERT(multi_lengths != NULL, "out of memory");
	if (multi_lengths == NULL)
	{
		return GEZ_UNKNOWN;
	}

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, QUAD_VERTEX_BUFFER);
		ASSERT(glGetError() == GL_NO_ERROR, "error");

	glBufferDataARB(GL_ARRAY_BUFFER_ARB,
					MAP_WIDTH * MAP_HEIGHT * sizeof(vertex) * 2 /* vertex & color data*/ + 
					MAP_WIDTH * MAP_HEIGHT * sizeof(text_coord),
					NULL,
					GL_STATIC_DRAW_ARB);
		ASSERT(glGetError() == GL_NO_ERROR, "error");

	glBufferSubDataARB(GL_ARRAY_BUFFER_ARB, 0, MAP_WIDTH * MAP_HEIGHT * sizeof(vertex), vertex_data);
		ASSERT(glGetError() == GL_NO_ERROR, "error");

	glBufferSubDataARB(GL_ARRAY_BUFFER_ARB, MAP_WIDTH * MAP_HEIGHT * sizeof(vertex), MAP_WIDTH * MAP_HEIGHT * sizeof(vertex), color_data);
		ASSERT(glGetError() == GL_NO_ERROR, "error");

	glBufferSubDataARB(GL_ARRAY_BUFFER_ARB, MAP_WIDTH * MAP_HEIGHT * sizeof(vertex) * 2, MAP_WIDTH * MAP_HEIGHT * sizeof(text_coord), texture_coords);
		ASSERT(glGetError() == GL_NO_ERROR, "error");

	glVertexPointer(3, GL_FLOAT, 0, BUFFER_OFFSET(0));
	glColorPointer(3, GL_FLOAT, 0, BUFFER_OFFSET(MAP_WIDTH * MAP_HEIGHT * sizeof(vertex)));
	glTexCoordPointer(2, GL_FLOAT, 0, BUFFER_OFFSET(MAP_WIDTH * MAP_HEIGHT * sizeof(vertex) * 2));

	glEnable(GL_VERTEX_ARRAY);
	glEnable(GL_COLOR_ARRAY);
	glEnable(GL_TEXTURE_COORD_ARRAY);

	// index buffer
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, QUAD_INDEX_BUFFER);
	  ASSERT(glGetError() == GL_NO_ERROR, "error");

	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, MAP_WIDTH * MAP_HEIGHT * sizeof(GLuint) * 12, NULL, GL_DYNAMIC_DRAW_ARB);
	  ASSERT(glGetError() == GL_NO_ERROR, "error");

	if (prefs.GET_INT_PARAM(PARAM_USE_MULTIDRAW_EXT))
	{
		console->Output("querying GL_EXT_multi_draw_arrays ");
		if (QueryExtension("GL_EXT_multi_draw_arrays"))
		{
			console->Output("\n");

			error = 0;

			GL_QUERY_PROC(glMultiDrawArraysEXT,   PFNGLMULTIDRAWARRAYSEXTPROC)
			GL_QUERY_PROC(glMultiDrawElementsEXT, PFNGLMULTIDRAWELEMENTSEXTPROC);

			if (!error)
			{
				multi_draw = true;
			}
			else
				multi_draw = false;
		}
		else
		{
			multi_draw = false;

			console->Output("not available\n");
		}
	}

	return GEZ_OK;
}

void QuadTreeRenderer::QuadTreeRender_TriangleFan_Recursion_ARB(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	int x, y, edge_len_x, edge_len_y;
	int MAPWIDTH_times_y,
		MAPWIDTH_times_y1,
		MAPWIDTH_times_y2;

	ASSERT(multi_indexes_ARB != NULL, "oops");
	ASSERT(multi_lengths != NULL, "oops, I did it again");
	ASSERT(index_pointer != NULL, "index pointer is NULL");

	edge_len_x = (x2 - x1);
	edge_len_y = (y2 - y1); // differ because of integer arithmetics

	x = x1 + edge_len_x / 2;
	y = y1 + edge_len_y / 2;

	MAPWIDTH_times_y = MAP_WIDTH * y;

	if ((quadtree_data[MAPWIDTH_times_y + x] == NODE_REFINE) && (edge_len_x > 3))
	{
		QuadTreeRender_TriangleFan_Recursion_ARB(x1,  y1,  x,  y);
		QuadTreeRender_TriangleFan_Recursion_ARB( x,  y1, x2,  y);
		QuadTreeRender_TriangleFan_Recursion_ARB( x,  y,  x2, y2);
		QuadTreeRender_TriangleFan_Recursion_ARB(x1,  y,  x,  y2);
	}
	else
	{
		if (quadtree_data[MAPWIDTH_times_y + x] == NODE_RENDER)
		{
			ASSERT(index_count < MAP_WIDTH * MAP_HEIGHT, "too many vertices");

			MAPWIDTH_times_y1 = MAP_WIDTH * y1;
			MAPWIDTH_times_y2 = MAP_WIDTH * y2;

			GLuint dummy;
			dummy = index_count;

			multi_indexes_ARB[fan_current] = index_count << 2;
			index_pointer[index_count++] = MAPWIDTH_times_y + x;
			index_pointer[index_count++] = MAP_WIDTH * y2 + x2; // bottom right

			// test whether coarser node to right
			if (x + edge_len_x < MAP_WIDTH)
			{
				if (quadtree_data[MAPWIDTH_times_y + x + edge_len_x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y + x2; // right
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y + x2; // right 
			}
	        
			index_pointer[index_count++] = MAPWIDTH_times_y1 + x2; // right top 

			// prevent cracking with upper neighbour 
			if (y - edge_len_y >= 0)
			{
				if (quadtree_data[MAP_WIDTH * (y - edge_len_y) + x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y1 +  x; // top 
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y1 +  x; // top 
			}

			index_pointer[index_count++] = MAPWIDTH_times_y1 + x1; // left top 

			// prevent cracking with left neighbour
			if (x - edge_len_x >= 0)
			{
				if (quadtree_data[MAPWIDTH_times_y + x - edge_len_x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y + x1; // left 
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y + x1; // left 
			}

			index_pointer[index_count++] = MAPWIDTH_times_y2 + x1; // left bottom 
			
			// prevent cracking with lower neighbour 
			if (y + edge_len_y < MAP_HEIGHT)
			{
				if (quadtree_data[MAP_WIDTH * (y + edge_len_y) + x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y2 +  x; // bottom 
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y2 +  x; // bottom 
			}

			index_pointer[index_count++] = MAPWIDTH_times_y2 + x2; // right bottom 

			multi_lengths[fan_current] = index_count - dummy;
			++fan_current;

			triangles_count += index_count - dummy - 2;
			vertices_count += index_count - dummy;
		}
	}
}

/* render quadtree with ARB_vendor extensions */
void QuadTreeRenderer::QuadTreeRender_ARB(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	GLuint i;

	index_count = 0;
    fan_current  = 0;

    /* create triangle fans */
	index_pointer = (GLuint *)glMapBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, GL_WRITE_ONLY);
	ASSERT(index_pointer != NULL, "couldn't map element array to client memory");
 
	QuadTreeRender_TriangleFan_Recursion_ARB(x1, y1, x2, y2);

	if (!glUnmapBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB))
	{
		WARN("couldn't unmap client memory");
		return ;
	}

	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, QUAD_INDEX_BUFFER);

	if (multi_draw)
	{
		glMultiDrawElementsEXT(GL_TRIANGLE_FAN, multi_lengths, GL_UNSIGNED_INT, (const GLvoid **)multi_indexes_ARB, fan_current);
	}
	else
	{
		for (i = 0; i < fan_current; i++)
		{
			glDrawElements(GL_TRIANGLE_FAN, multi_lengths[i], GL_UNSIGNED_INT, BUFFER_OFFSET(multi_indexes_ARB[i])); // values are * sizeof(GLuint) by recursion function
		}
	}



	return;
}
