/*

  the implementation of the Console class

*/

#include "Console.h"
#include "Assert.h"
#include "stdio.h"
#include "stdarg.h"
#include "preferences.h"

#include "application.h"

extern Application *theApp;

Console::Console()
{
	gui_init = false;
	log_name = NULL;

}

Console::~Console()
{
	if (gui_init)
	{
		glDeleteLists(base, 96);
	}
}

bool Console::Initialize(char *log_file)
{
  FILE *file;

  log_name = log_file;

  file = fopen(log_name, "w");
  ASSERT(file != NULL, "error creating file");

  if (file == NULL)
  {
	  return false;
  }

  fclose(file);
  
  return true;
}

void Console::Separator(bool bold)
{
	if (bold)
		Output("==================================================\n");
	else
		Output("--------------------------------------------------\n");
}

#include <math.h>

bool Console::init_gui()
{
	int i;

	int width, height;

	ASSERT(theApp != NULL, "application is not set");

	width  = theApp->prefs.GET_INT_PARAM(PARAM_WIDTH);
	height = theApp->prefs.GET_INT_PARAM(PARAM_HEIGHT);

	gui_init = false;

	QueryPerformanceFrequency((LARGE_INTEGER *)&delete_interval);
	delete_interval *= 5;

	y_offset = (1.0f / float(height)) * 16.0f;

	GuiResetYPos();
	pos = 0;

	for (i = 0; i < 8; i++)
	{
		con_buf[i][0] = '\0';
		con_times[i] = 0;
	}

	HFONT	font;

	font = CreateFont(  -16,							// Height Of Font
						0,								// Width Of Font
						0,								// Angle Of Escapement
						0,								// Orientation Angle
						FW_BOLD,						// Font Weight
						FALSE,							// Italic
						FALSE,							// Underline
						FALSE,							// Strikeout
						ANSI_CHARSET,					// Character Set Identifier
						OUT_TT_PRECIS,					// Output Precision
						CLIP_DEFAULT_PRECIS,			// Clipping Precision
						ANTIALIASED_QUALITY,			// Output Quality
						FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
						"Verdana");						// Font Name

	if (SelectObject(theApp->hdc, font) ==  NULL)		// selects the font
	{
		Output("WARNING: couldn't create font");
		return false;
	}

	base = glGenLists(96);
	if (wglUseFontBitmaps(theApp->hdc, 32, 96, base) != TRUE)			// Builds 96 Characters Starting At Character 32
	{
		Output("WARNING: couldn't create font");
		return false;
	}

	gui_init = true;

	return true;
}

void Console::ToVaList(const char *format, va_list va)
{
	FILE *file;

	ASSERT(log_name != NULL, "log file name not set!")

	file = fopen(log_name, "a");
	ASSERT(file != NULL, "error creating file");

	if (file != NULL)
	{
		vfprintf(file, format, va);
		fclose(file);
	}
}

void Console::Output(const char *format, ...)
{
  va_list va;
  
  va_start(va, format);
	ToVaList(format, va);
  va_end(va);
}

void Console::ScrOutput(const char *string, ...)
{
	va_list va;

	va_start(va, string);
		vsprintf(con_buf[pos], string, va);
		ToVaList(string, va);
	va_end(va);

	QueryPerformanceCounter((LARGE_INTEGER *)&con_times[pos++]);

	if (pos >= 8)
	{
		pos = 0;
	}
}

void Console::ShowConsole(bool reset_y_pos)
{
	int i;

	if (reset_y_pos) GuiResetYPos();

	QueryPerformanceCounter((LARGE_INTEGER *)&time);

	i = pos;
	do
	{
		if (con_buf[i][0] != '\0')
		{
			ScrPrint(con_buf[i]);

			if (time > (con_times[i] + delete_interval))
			{
				con_buf[i][0] = '\0';
				con_times[i]  = 0;
			}
		}


		if (++i >= 8)
		{
			i = 0;
		}
	} while (i != pos);
}

void Console::ScrPrint(const char *string) // CAUTION: destroys ModelView Matrix
{
	ASSERT(gui_init, "gui is not intialized");

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glColor3f(1.0f, 1.0f, 1.0f);
	glTranslatef(0.0f,0.0f, -1.0f);
	glRasterPos2f(-0.55f, y_pos);
	y_pos -= y_offset;

	glListBase(base - 32);
	  
	glCallLists(strlen(string), GL_UNSIGNED_BYTE, string);
}
