//#include "stdafx.h"

#include <stdio.h>

#include "loadbmp.h"
#include "error.h"

BITMAP *LoadMapFromBitmapFile(char *name, BITMAP *bm)
{
	//file handle used in all file operations
    FILE *in;
    RGBQUAD *colours;
    BYTE *pixelData;
    LONG width,height;
    WORD bpp;
	int numColours;

    //open the file for reading in binary mode    
    in=fopen(name,"rb");

	if (in == NULL) return NULL;

	BITMAPFILEHEADER bmfh;

	fread(&bmfh,sizeof(BITMAPFILEHEADER),1,in);

	BITMAPINFOHEADER bmih;

	fread(&bmih,sizeof(BITMAPINFOHEADER),1,in);	//set the number of colours
	
	numColours=1 << bmih.biBitCount;

	//load the palette for 8 bits per pixel
	if(bmih.biBitCount == 8) {
		colours=new RGBQUAD[numColours];
		fread(colours,sizeof(RGBQUAD),numColours,in);
	}
	else
	{
		fclose(in);
		FAIL("can only load 256 colors bitmaps as map");
		return NULL;
	}

	BYTE *tempPixelData;
	int size;

	size=bmfh.bfSize-bmfh.bfOffBits;
	width = bmih.biWidth;
	height = bmih.biHeight;
	bpp = bmih.biBitCount;

	tempPixelData=new BYTE[size];

	if(tempPixelData==NULL) {
		fclose(in);
		return NULL;
	}
	if (fread(tempPixelData,sizeof(BYTE),size,in) != size)
	{
		fclose(in);
		delete[] tempPixelData;
		return NULL;
	};

	//byteWidth is the width of the actual image in bytes
	//padWidth is the width of the image plus the extra padding
	LONG byteWidth,padWidth;

	//initially set both to the width of the image
	byteWidth=padWidth=(LONG)((float)width*(float)bpp/8.0);

	//add any extra space to bring each line to a DWORD boundary
	while(padWidth%4!=0) {
		padWidth++;
	}

	DWORD diff;
	int offset;

	height=bmih.biHeight;
	//set diff to the actual image size(no padding)
	diff=height*byteWidth;
	//allocate memory for the image
	pixelData=new BYTE[diff];
	if(pixelData==NULL) {
		fclose(in);
		return NULL;
	}
	//bitmap is inverted, so the padding needs to be removed
	//and the image reversed
	//Here you can start from the back of the file or the front,
	//after the header.  The only problem is that some programs
	//will pad not only the data, but also the file size to
	//be divisible by 4 bytes.
	if (height > 0)
	{
//		int j = diff - 1;
		int bpp_8 = bpp / 8;
		offset = padWidth - byteWidth;

		int i, j;
		for(j = i = 0; i < size, j <= diff - 1; i += bpp_8)
		{
			if ((i + offset) %padWidth == 0) {
				i += offset;
			}; 

		  *(pixelData + j)= *(tempPixelData + i); // kann man optimieren mit memcpy

		  ++j;
		}
	}
	//the image is not reversed.  Only the padding needs to be removed.
	else {
		height=height*-1;
		offset=0;
		do {
			memcpy((pixelData+(offset*byteWidth)),
				(tempPixelData+(offset*padWidth)),
					byteWidth);
			offset++;
		} while(offset<height);
	}

	bm->bmBits = pixelData;
	bm->bmType = 0;
	bm->bmBitsPixel = bpp;
	bm->bmHeight = height;
	bm->bmWidth = width;
	bm->bmWidthBytes = byteWidth; // FIXME das stimmt nicht
	bm->bmPlanes = bmih.biPlanes;

	delete[] tempPixelData;

	return bm;
}

BITMAP *LoadTextureFromBitmapFile(char *name, BITMAP *bm)
{
	//file handle used in all file operations
    FILE *in;
    RGBQUAD *colours;
    BYTE *pixelData;
    LONG width,height;
    WORD bpp;
	int numColours;

    //open the file for reading in binary mode    
    in=fopen(name,"rb");

	if (in == NULL) return NULL;

	BITMAPFILEHEADER bmfh;

	fread(&bmfh,sizeof(BITMAPFILEHEADER),1,in);

	BITMAPINFOHEADER bmih;

	fread(&bmih,sizeof(BITMAPINFOHEADER),1,in);	//set the number of colours
	
	numColours=1 << bmih.biBitCount;

	//load the palette for 8 bits per pixel
	if(bmih.biBitCount == 8) {
		colours=new RGBQUAD[numColours];
		fread(colours,sizeof(RGBQUAD),numColours,in);
	}

	BYTE *tempPixelData;
	int size;

	size=bmfh.bfSize-bmfh.bfOffBits;
	width = bmih.biWidth;
	height = bmih.biHeight;
	bpp = bmih.biBitCount;

	tempPixelData=new BYTE[size];

	if(tempPixelData==NULL) {
		fclose(in);
		return NULL;
	}
	if (fread(tempPixelData,sizeof(BYTE),size,in) != size)
	{
		fclose(in);
		delete[] tempPixelData;
		return NULL;
	};

	//byteWidth is the width of the actual image in bytes
	//padWidth is the width of the image plus the extra padding
	LONG byteWidth,padWidth;

	//initially set both to the width of the image
	byteWidth=padWidth=(LONG)((float)width*(float)bpp/8.0);

	//add any extra space to bring each line to a DWORD boundary
	while(padWidth%4!=0) {
		padWidth++;
	}

	DWORD diff;
	int offset;

	height=bmih.biHeight;
	//set diff to the actual image size(no padding)
	diff=height*byteWidth;
	//allocate memory for the image

	if (bpp == 8) diff *=3;
	pixelData = new BYTE[diff];

	if(pixelData==NULL) {
		fclose(in);
		return NULL;
	}
	//bitmap is inverted, so the padding needs to be removed
	//and the image reversed
	//Here you can start from the back of the file or the front,
	//after the header.  The only problem is that some programs
	//will pad not only the data, but also the file size to
	//be divisible by 4 bytes.
	if (height > 0)
	{
		int bpp_8 = bpp / 8;
//		int j = diff - 3;
		offset = padWidth - byteWidth;

		int i, j;
		for(j = i = 0; i < size, j <= diff - 3; i += (bpp_8))
		{
			if ((i + offset) %padWidth == 0) {
				i += offset;
			}; 

			if (bpp_8 == 1)
			{
				*(pixelData + j) = colours[*(tempPixelData + i + (bpp_8) - 1)].rgbRed;
				*(pixelData + j + 1) = colours[*(tempPixelData + i + (bpp_8) - 1)].rgbGreen;
				*(pixelData + j + 2) = colours[*(tempPixelData + i + (bpp_8) - 1)].rgbBlue;
			}
			else
			{
				*(pixelData + j)	 = *(tempPixelData + i + 2);
				*(pixelData + j + 1) = *(tempPixelData + i + 1);
				*(pixelData + j + 2) = *(tempPixelData + i);
			}

			j += 3;
		}
	}
	//the image is not reversed.  Only the padding needs to be removed.
	else {
		height=height*-1;
		offset=0;
		do {
			memcpy((pixelData+(offset*byteWidth)),
				(tempPixelData+(offset*padWidth)),
					byteWidth);
			offset++;
		} while(offset<height);
	}

	bm->bmBits = pixelData;
	bm->bmType = 0;
	bm->bmBitsPixel = bpp;
	bm->bmHeight = height;
	bm->bmWidth = width;
	bm->bmWidthBytes = byteWidth; // FIXME das stimmt nicht
	bm->bmPlanes = bmih.biPlanes;

	delete[] tempPixelData;

	return bm;
}
