#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#define DIRECTINPUT_VERSION 0x0800
#include <dinput.h>

#include "main.h"
#include "assert.h"
#include "Application.h"
#include "preferences.h"

#include <stdio.h>

/*  Declare Windows procedure  */

bool benchmark;

/*  Make the class name into a global variable  */
char szClassName[ ] = CLASSNAME;

/* define the global Application Object */
static Application *App; 

int WINAPI WinMain (HINSTANCE hThisInstance,
                    HINSTANCE hPrevInstance,
                    LPSTR lpszArgument,
                    int nFunsterStil)

{
    GEZERROR result;
    MSG messages;            /* Here messages to the application are saved */

    /* create the application object */
    App = new Application();
    ASSERT(App != NULL, "App is NULL")
    
    /* initialize application */
    if ((result = App->Initialize(hThisInstance)) != GEZ_OK)
    {
      return false;
    }

    /* Make the window visible on the screen */
    ShowWindow (App->hwnd, nFunsterStil);

	benchmark = App->prefs.GET_INT_PARAM(PARAM_BENCHMARK) != 0;

    /* Run the message loop. It will run until GetMessage() returns 0 */
    ZeroMemory(&messages, sizeof(messages));
    while (messages.message != WM_QUIT)
    {
		if (PeekMessage (&messages, NULL, 0, 0, PM_REMOVE) != 0)
		{
			/* Translate virtual-key messages into character messages */
			TranslateMessage(&messages);
			/* Send message to WindowProcedure */
			DispatchMessage(&messages);
		}
		else
		{
			/* TODO (gez#1#): reduce function call overhead by directly calling 
			render routines, alternatively or multi-threading */
			App->Render();

			if (App->bActive)
			{

				/* we have input data */
				if (!benchmark && WaitForSingleObject(App->hMouseEvent, 0) == WAIT_OBJECT_0)
				{
					DIDEVICEOBJECTDATA didod;
					DWORD count, x, y;
					x = y = 0;

					count = 1;
					while (count)
					{
						count = 1;
						if (FAILED(App->lpDIMouse->GetDeviceData(sizeof(didod), &didod, &count, 0)))
						{
							App->lpDIMouse->Acquire();
							App->lpDIMouse->GetDeviceData(sizeof(didod), &didod, &count, 0);
						};

						switch (didod.dwOfs)
						{
							case DIMOFS_X:
								x = didod.dwData;
								App->MousePos(x, 0);
								break;
							case DIMOFS_Y:
								y = didod.dwData;
								App->MousePos(0, y);
								break;
						}
					}
				}
			}
		}

		if (benchmark)
		{
			if (App->frames_rendered >= App->camera.map->height * 8)
			{
				__int64	ft_creation,
						ft_exit,
						ft_kernel,
						ft_user;
				SYSTEMTIME  time;

				char msg[50];

				HANDLE h;

				h = OpenThread(THREAD_QUERY_INFORMATION, FALSE, App->threadID);
				if (!GetThreadTimes(h, (LPFILETIME)&ft_creation, (LPFILETIME)&ft_exit, (LPFILETIME)&ft_kernel, (LPFILETIME)&ft_user))
				{
					MessageBox(NULL, "error retrieving needed time", "error", MB_OK | MB_SYSTEMMODAL);
				}
				else
				{
					ft_user = ft_user + ft_kernel;
					if (!FileTimeToSystemTime((LPFILETIME)&ft_user, &time))
					{
						MessageBox(NULL, "error calculating needed time", "error", MB_OK | MB_SYSTEMMODAL);
					}
					else
					{
						sprintf(msg, "%02d:%02d.%03d taken", time.wMinute, time.wSecond, time.wMilliseconds);
						MessageBox(NULL, msg, "Time Taken", MB_OK | MB_SYSTEMMODAL);
					}
				}
				CloseHandle(h);

				benchmark = false;
				PostMessage(App->hwnd, WM_QUIT, 0, 0);
			}
		}
    }

    /* The program return-value is 0 - The value that PostQuitMessage() gave */
    return messages.wParam;
}



